/* Program header manipulation tool
   Copyright (C) 2010-2014 Free Software Foundation, Inc.

   Parts of this file are taken from binutils/elfcomm.c in the binutils-gdb
   project.

   Originally developed by Eric Youngdale <eric@andante.jic.com>
   Modifications by Nick Clifton <nickc@redhat.com>
   pheader.c utility by Matthew Fortune <matthew.fortune@imgtec.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street - Fifth Floor, Boston, MA
   02110-1301, USA.  */

#include <elf.h>
#include "external.h"
#include <stdio.h>
#include <stdlib.h>

#define BYTE_GET(field)              byte_get (field, sizeof (field))
#define BYTE_PUT(field, val)    byte_put (field, val, sizeof (field))

void
byte_put_little_endian (unsigned char * field, uint64_t value, int size)
{
  switch (size)
    {
    case 8:
      field[7] = (((value >> 24) >> 24) >> 8) & 0xff;
      field[6] = ((value >> 24) >> 24) & 0xff;
      field[5] = ((value >> 24) >> 16) & 0xff;
      field[4] = ((value >> 24) >> 8) & 0xff;
      /* Fall through.  */
    case 4:
      field[3] = (value >> 24) & 0xff;
      /* Fall through.  */
    case 3:
      field[2] = (value >> 16) & 0xff;
      /* Fall through.  */
    case 2:
      field[1] = (value >> 8) & 0xff;
      /* Fall through.  */
    case 1:
      field[0] = value & 0xff;
      break;

    default:
      error ("Unhandled data length: %d\n", size);
      abort ();
    }
}

void
byte_put_big_endian (unsigned char * field, uint64_t value, int size)
{
  switch (size)
    {
    case 8:
      field[7] = value & 0xff;
      field[6] = (value >> 8) & 0xff;
      field[5] = (value >> 16) & 0xff;
      field[4] = (value >> 24) & 0xff;
      value >>= 16;
      value >>= 16;
      /* Fall through.  */
    case 4:
      field[3] = value & 0xff;
      value >>= 8;
      /* Fall through.  */
    case 3:
      field[2] = value & 0xff;
      value >>= 8;
      /* Fall through.  */
    case 2:
      field[1] = value & 0xff;
      value >>= 8;
      /* Fall through.  */
    case 1:
      field[0] = value & 0xff;
      break;

    default:
      error ("Unhandled data length: %d\n", size);
      abort ();
    }
}

void (*byte_put) (unsigned char *, uint64_t, int) = byte_put_little_endian;


uint64_t
byte_get_little_endian (unsigned char *field, int size)
{
  switch (size)
    {
    case 1:
      return *field;

    case 2:
      return  ((unsigned int) (field[0]))
	|    (((unsigned int) (field[1])) << 8);

    case 3:
      return  ((unsigned long) (field[0]))
	|    (((unsigned long) (field[1])) << 8)
	|    (((unsigned long) (field[2])) << 16);

    case 4:
      return  ((unsigned long) (field[0]))
	|    (((unsigned long) (field[1])) << 8)
	|    (((unsigned long) (field[2])) << 16)
	|    (((unsigned long) (field[3])) << 24);

    case 5:
	return  ((uint64_t) (field[0]))
	  |    (((uint64_t) (field[1])) << 8)
	  |    (((uint64_t) (field[2])) << 16)
	  |    (((uint64_t) (field[3])) << 24)
	  |    (((uint64_t) (field[4])) << 32);

    case 6:
	return  ((uint64_t) (field[0]))
	  |    (((uint64_t) (field[1])) << 8)
	  |    (((uint64_t) (field[2])) << 16)
	  |    (((uint64_t) (field[3])) << 24)
	  |    (((uint64_t) (field[4])) << 32)
	  |    (((uint64_t) (field[5])) << 40);

    case 7:
	return  ((uint64_t) (field[0]))
	  |    (((uint64_t) (field[1])) << 8)
	  |    (((uint64_t) (field[2])) << 16)
	  |    (((uint64_t) (field[3])) << 24)
	  |    (((uint64_t) (field[4])) << 32)
	  |    (((uint64_t) (field[5])) << 40)
	  |    (((uint64_t) (field[6])) << 48);

    case 8:
	return  ((uint64_t) (field[0]))
	  |    (((uint64_t) (field[1])) << 8)
	  |    (((uint64_t) (field[2])) << 16)
	  |    (((uint64_t) (field[3])) << 24)
	  |    (((uint64_t) (field[4])) << 32)
	  |    (((uint64_t) (field[5])) << 40)
	  |    (((uint64_t) (field[6])) << 48)
	  |    (((uint64_t) (field[7])) << 56);

    default:
      error ("Unhandled data length: %d\n", size);
      abort ();
    }
}

uint64_t
byte_get_big_endian (unsigned char *field, int size)
{
  switch (size)
    {
    case 1:
      return *field;

    case 2:
      return ((unsigned int) (field[1])) | (((int) (field[0])) << 8);

    case 3:
      return ((unsigned long) (field[2]))
	|   (((unsigned long) (field[1])) << 8)
	|   (((unsigned long) (field[0])) << 16);

    case 4:
      return ((unsigned long) (field[3]))
	|   (((unsigned long) (field[2])) << 8)
	|   (((unsigned long) (field[1])) << 16)
	|   (((unsigned long) (field[0])) << 24);

    case 5:
	return ((uint64_t) (field[4]))
	  |   (((uint64_t) (field[3])) << 8)
	  |   (((uint64_t) (field[2])) << 16)
	  |   (((uint64_t) (field[1])) << 24)
	  |   (((uint64_t) (field[0])) << 32);

    case 6:
	  /* Although we are extracting data from an 8 byte wide field,
	     we are returning only 4 bytes of data.  */
	  field += 2;
	  return ((unsigned long) (field[3]))
	    |   (((unsigned long) (field[2])) << 8)
	    |   (((unsigned long) (field[1])) << 16)
	    |   (((unsigned long) (field[0])) << 24);

    case 7:
	return ((uint64_t) (field[6]))
	  |   (((uint64_t) (field[5])) << 8)
	  |   (((uint64_t) (field[4])) << 16)
	  |   (((uint64_t) (field[3])) << 24)
	  |   (((uint64_t) (field[2])) << 32)
	  |   (((uint64_t) (field[1])) << 40)
	  |   (((uint64_t) (field[0])) << 48);

    case 8:
	return ((uint64_t) (field[7]))
	  |   (((uint64_t) (field[6])) << 8)
	  |   (((uint64_t) (field[5])) << 16)
	  |   (((uint64_t) (field[4])) << 24)
	  |   (((uint64_t) (field[3])) << 32)
	  |   (((uint64_t) (field[2])) << 40)
	  |   (((uint64_t) (field[1])) << 48)
	  |   (((uint64_t) (field[0])) << 56);

    default:
      error ("Unhandled data length: %d\n", size);
      abort ();
    }
}

uint64_t (*byte_get) (unsigned char *, int) = byte_get_little_endian;

void
usage (void)
{
  printf ("Usage: pheader <filename> corrupt\n");
  printf ("                          delete\n");
  printf ("                          badflags2\n");
  printf ("                          badase\n");
  printf ("                          msaase\n");
  printf ("                          isaext\n");
  printf ("                          setfpabi <value> [<oddspreg>]\n");
  exit (1);
}

int
main (int argc, const char **argv)
{
  Elf32_External_Ehdr header32_ex;
  Elf32_External_Phdr pheader32_ex;
  Elf64_External_Ehdr header64_ex;
  Elf64_External_Phdr pheader64_ex;
  Elf_External_ABIFlags_v0 abiflags_ex;
  FILE *fp;
  int i;
  int big_endian;
  char fheader[16];
  int pheader_offset;
  int abiflags_offset;
  short int pheader_count;
  short int pheader_entrysize;

  if (argc < 3)
    usage ();

  fp = fopen (argv[1], "r+");

  if (fp == NULL)
    {
      printf ("Invalid file: %s\n", argv[1]);
      exit (1);
    }

  if (fread (&fheader, sizeof (char), 16, fp) != 16)
    {
      printf ("Can't read elf header\n");
      exit (1);
    }

  if (fheader[EI_MAG0] != ELFMAG0
      || fheader[EI_MAG1] != ELFMAG1
      || fheader[EI_MAG2] != ELFMAG2
      || fheader[EI_MAG3] != ELFMAG3)
    {
      printf ("Invalid ELF file\n");
      exit (1);
    }

  /* Set the file offset to the start so we correctly read the elf header.  */
  rewind (fp);

  if (fheader[EI_CLASS] == ELFCLASS32)
    {
      if (fread (&header32_ex, sizeof (Elf32_External_Ehdr), 1, fp) != 1)
	{
	  printf ("ELF header read error\n");
	  exit (1);
	}
      if (header32_ex.e_ident[EI_DATA] == ELFDATA2MSB)
	{
	  byte_put = byte_put_big_endian;
	  byte_get = byte_get_big_endian;
	}
      pheader_offset = BYTE_GET (header32_ex.e_phoff);
      pheader_count = BYTE_GET (header32_ex.e_phnum);
      pheader_entrysize = BYTE_GET (header32_ex.e_phentsize);
    }
  else if (fheader[EI_CLASS] == ELFCLASS64)
    {
      if (fread (&header64_ex, sizeof (Elf64_Ehdr), 1, fp) != 1)
	{
	  printf ("ELF header read error\n");
	  exit (1);
	}

      if (header64_ex.e_ident[EI_DATA] == ELFDATA2MSB)
	{
	  byte_put = byte_put_big_endian;
	  byte_get = byte_get_big_endian;
	}
      pheader_offset = BYTE_GET (header64_ex.e_phoff);
      pheader_count = BYTE_GET (header64_ex.e_phnum);
      pheader_entrysize = BYTE_GET (header64_ex.e_phentsize);
    }
  else
    {
      printf ("Invalid ELF file\n");
      exit (1);
    }

  if (fseek (fp, pheader_offset, SEEK_SET) == -1)
    {
      printf ("Can't find program headers\n");
      exit (1);
    }

  for (i = 0; i < pheader_count; i++)
    {
      if (fheader[EI_CLASS] == ELFCLASS32)
	{
	  if (fread (&pheader32_ex, pheader_entrysize, 1, fp) != 1)
	    {
	      printf ("Program header read error elf32 %d\n", pheader_entrysize);
	      exit (1);
	    }
	  if (BYTE_GET (pheader32_ex.p_type) == 0x70000003)
	    {
	      abiflags_offset = BYTE_GET (pheader32_ex.p_offset);
	      if (BYTE_GET (pheader32_ex.p_filesz) != sizeof(abiflags_ex))
		{
		  printf ("abiflags size is incorrect\n");
		  exit(1);
		}
	      break;
	    }
	}
      else if (fheader[EI_CLASS] == ELFCLASS64)
	{
	  if (fread (&pheader64_ex, pheader_entrysize, 1, fp) != 1)
	    {
	      printf ("Program header read error elf64\n");
	      exit (1);
	    }
	  if (BYTE_GET (pheader64_ex.p_type) == 0x70000003)
	    {
	      abiflags_offset = BYTE_GET (pheader64_ex.p_offset);
	      if (BYTE_GET (pheader64_ex.p_filesz) != sizeof(abiflags_ex))
		{
		  printf ("abiflags size is incorrect\n");
		  exit(1);
		}
	      break;
	    }
	}
    }

  if (i == pheader_count)
    {
      printf ("Can't find MIPS ABIflags\n");
      exit (1);
    }

  if (strcmp (argv[2], "delete") == 0)
    {
      if (fheader[EI_CLASS] == ELFCLASS32)
	BYTE_PUT (pheader32_ex.p_type, PT_NULL);
      else
	BYTE_PUT (pheader64_ex.p_type, PT_NULL);
    }
  else if (strcmp (argv[2], "corrupt") == 0)
    {
      if (fheader[EI_CLASS] == ELFCLASS32)
	{
	  BYTE_PUT (pheader32_ex.p_filesz, 0x10);
	  BYTE_PUT (pheader32_ex.p_memsz, 0x10);
	}
      else
	{
	  BYTE_PUT (pheader64_ex.p_filesz, 0x10);
	  BYTE_PUT (pheader64_ex.p_memsz, 0x10);
	}
    }

  if (fseek (fp, -pheader_entrysize, SEEK_CUR) == -1)
    {
      printf ("Can't find MIPS ABI flags headers\n");
      exit (1);
    }

  if (fheader[EI_CLASS] == ELFCLASS32)
    {
      if (fwrite (&pheader32_ex, pheader_entrysize, 1, fp) != 1)
	{
	  printf ("Program headers write error\n");
	  exit (1);
	}
    }
  else
    {
      if (fwrite (&pheader64_ex, pheader_entrysize, 1, fp) != 1)
	{
	  printf ("Program headers write error\n");
	  exit (1);
	}
    }

  if (fseek (fp, abiflags_offset, SEEK_SET) == -1)
    {
      printf ("Can't find abiflags\n");
      exit (1);
    }

  if (fread (&abiflags_ex, sizeof (abiflags_ex), 1, fp) != 1)
    {
      printf ("abiflags read error\n");
      exit (1);
    }

  if (strcmp (argv[2], "badase") == 0
      || strcmp (argv[2], "msaase") == 0)
    {
#define AFL_ASE_MSA          0x00000200
      uint32_t ases = BYTE_GET (abiflags_ex.ases);
      if (strcmp (argv[2], "badase") == 0)
	ases |= 0x80000000;
      else
	ases |= AFL_ASE_MSA;

      BYTE_PUT (abiflags_ex.ases, ases);
    }

  if (strcmp (argv[2], "isaext") == 0)
    {
      uint32_t ext = BYTE_GET (abiflags_ex.isa_ext);
      ext = 102;

      BYTE_PUT (abiflags_ex.isa_ext, ext);
    }


  if (strcmp (argv[2], "badflags2") == 0)
    {
      uint32_t flags2 = BYTE_GET (abiflags_ex.flags2);
      flags2 |= 1;

      BYTE_PUT (abiflags_ex.flags2, flags2);
    }

  if (strcmp (argv[2], "setfpabi") == 0)
    {
      int newval;
      if (argc < 4)
	usage();

      newval = atoi (argv[3]);

      BYTE_PUT (abiflags_ex.fp_abi, newval);

      if (argc > 4)
	{
	  uint32_t flags1 = BYTE_GET (abiflags_ex.flags1);
	  if (atoi (argv[4]) == 1)
	    flags1 |= 0x1;
	  else
	    flags1 &= ~0x1;
	  BYTE_PUT (abiflags_ex.flags1, flags1);
	}
    }
  if (fseek (fp, abiflags_offset, SEEK_SET) == -1)
    {
      printf ("Can't find abiflags\n");
      exit (1);
    }

  if (fwrite (&abiflags_ex, sizeof (abiflags_ex), 1, fp) != 1)
    {
      printf ("Program headers write error\n");
      exit (1);
    }

  if (fclose (fp) == EOF)
    {
      printf ("File closing error\n");
      exit (1);
    }

  return 0;
}
